package WebService::HIBP;

use strict;
use warnings;
use JSON();
use URI::Escape();
use LWP::UserAgent();
use Digest::SHA();
use WebService::HIBP::Breach();
use WebService::HIBP::Paste();

our $VERSION = '0.01';

sub _LENGTH_OF_PASSWORD_PREFIX { return 5; }

sub new {
    my ($class) = @_;
    my $self = {};
    bless $self, $class;
    $self->{url}          = 'https://haveibeenpwned.com/api/v2/';
    $self->{password_url} = 'https://api.pwnedpasswords.com/range/';
    $self->{ua}           = LWP::UserAgent->new( agent => 'WebService::HIBP ' );
    $self->{ua}->env_proxy();
    return $self;
}

sub data_classes {
    my ($self)   = @_;
    my $url      = $self->{url} . 'dataclasses';
    my $response = $self->{ua}->get($url);
    if ( $response->is_success() ) {
        my $json = JSON::decode_json( $response->decoded_content() );
        my @classes;
        foreach my $class ( @{$json} ) {
            push @classes, $class;
        }
        return @classes;
    }
    else {
        Carp::croak( "Failed to retrieve $url:" . $response->status_line() );
    }
}

sub breach {
    my ( $self, $name ) = @_;
    my $url      = $self->{url} . 'breach/' . $name;
    my $response = $self->{ua}->get($url);
    if ( $response->is_success() ) {
        my $json = JSON::decode_json( $response->decoded_content() );
        return WebService::HIBP::Breach->new( %{$json} );
    }
    else {
        Carp::croak( "Failed to retrieve $url:" . $response->status_line() );
    }
}

sub pastes {
    my ( $self, $account ) = @_;
    my $url =
      $self->{url} . 'pasteaccount/' . URI::Escape::uri_escape($account);
    my $response = $self->{ua}->get($url);
    if ( $response->is_success() ) {
        my $json = JSON::decode_json( $response->decoded_content() );
        my @pastes;
        foreach my $paste ( @{$json} ) {
            push @pastes, WebService::HIBP::Paste->new( %{$paste} );
        }
        return @pastes;
    }
    else {
        Carp::croak( "Failed to retrieve $url:" . $response->status_line() );
    }
}

sub breaches {
    my ( $self, %parameters ) = @_;
    my $url = $self->{url} . 'breaches';
    if ( $parameters{domain} ) {
        $url .= '?domain=' . URI::Escape::uri_escape( $parameters{domain} );
    }
    my $response = $self->{ua}->get($url);
    if ( $response->is_success() ) {
        my $json = JSON::decode_json( $response->decoded_content() );
        my @breaches;
        foreach my $breach ( @{$json} ) {
            push @breaches, WebService::HIBP::Breach->new( %{$breach} );
        }
        return @breaches;
    }
    else {
        Carp::croak( "Failed to retrieve $url:" . $response->status_line() );
    }
}

sub account {
    my ( $self, $account, %parameters ) = @_;
    my $url =
      $self->{url} . 'breachedaccount/' . URI::Escape::uri_escape($account);
    my @filters;
    if ( $parameters{unverified} ) {
        push @filters, 'includeUnverified=true';
    }
    if ( $parameters{truncate} ) {
        push @filters, 'truncateResponse=true';
    }
    if ( $parameters{domain} ) {
        push @filters,
          'domain=' . URI::Escape::uri_escape( $parameters{domain} );
    }
    if (@filters) {
        $url .= q[?] . join q[&], @filters;
    }
    my $response = $self->{ua}->get($url);
    if ( $response->is_success() ) {
        my $json = JSON::decode_json( $response->decoded_content() );
        my @breaches;
        foreach my $breach ( @{$json} ) {
            push @breaches, WebService::HIBP::Breach->new( %{$breach} );
        }
        return @breaches;
    }
    else {
        Carp::croak( "Failed to retrieve $url:" . $response->status_line() );
    }
}

sub password {
    my ( $self, $password ) = @_;
    my $sha1 = uc Digest::SHA::sha1_hex($password);
    my $url = $self->{password_url} . substr $sha1, 0,
      _LENGTH_OF_PASSWORD_PREFIX();
    my $response = $self->{ua}->get($url);
    if ( $response->is_success() ) {
        my $remainder = substr $sha1, _LENGTH_OF_PASSWORD_PREFIX();
        foreach my $line ( split /\r\n/smx, $response->decoded_content() ) {
            my ( $pwned, $count ) = split /:/smx, $line;
            if ( $pwned eq $remainder ) {
                return $count;
            }
        }
        return 0;
    }
    else {
        Carp::croak( "Failed to retrieve $url:" . $response->status_line() );
    }
}

1;    # End of WebService::HIBP
__END__

=head1 NAME

WebService::HIBP - An interface to the Have I Been Pwned webservice at haveibeenpwned.com

=head1 VERSION

Version 0.01

=head1 SYNOPSIS

Check the security of your accounts/email addresses and passwords

    use WebService::HIBP();
    use IO::Prompt();

    my $hibp = WebService::HIBP->new();
    my $new_password = IO::Prompt::prompt(-echo => q[*], 'Enter your new password:');
    my $count = $hibp->password($new_password);
    if ($count == 0) {
    } elsif ($count <= 2) {
       warn "This password has been found in a data breach\n";
    } elsif ($count) {
       die "This password is too insecure\n";
    }

=head1 DESCRIPTION

This is a client module for the L<https://haveibeenpwned.com/api/v2/> API, which provides a searchable interface to account/password breaches and pastes on sites such as pastebin.com

=head1 SUBROUTINES/METHODS

=head2 new

a new C<WebService::HIBP> object, ready to check how bad the pwnage is.

=head2 password

The L<Pwned Passwords API|https://haveibeenpwned.com/API/v2#PwnedPasswords> has more than half a billion passwords which have previously been exposed in data breaches. The service is detailed in the L<launch blog post|https://www.troyhunt.com/introducing-306-million-freely-downloadable-pwned-passwords/> then L<further expanded on with the release of version 2|https://www.troyhunt.com/ive-just-launched-pwned-passwords-version-2>. The entire data set is L<both downloadable and searchable online via the Pwned Passwords page|https://haveibeenpwned.com/Passwords>.

Each password is stored as a SHA-1 hash of a UTF-8 encoded password. The downloadable source data delimits the full SHA-1 hash and the password count with a colon (:) and each line with a CRLF.

In order to protect the value of the source password being searched for, this method implements a k-Anonymity model that searches for a password by partial hash. This method sends the first 5 characters of a SHA-1 password hash to be passed to the L<Pwned Passwords API|https://haveibeenpwned.com/API/v2#PwnedPasswords>.

The L<Pwned Passwords API|https://haveibeenpwned.com/API/v2#PwnedPasswords> will respond with a list of the suffix of every hash beginning with the specified prefix, followed by a count of how many times it appears in the data set. This method then searches the results of the response for the presence of their source hash and if not found, the password does not exist in the data set.  

It returns the count of how many times it appears in the data set or "0" if it dosen't appear.

    use WebService::HIBP();
    use IO::Prompt();

    my $hibp = WebService::HIBP->new();
    my $new_password = IO::Prompt::prompt(-echo => q[*], 'Enter your new password:');
    my $count = $hibp->password($new_password);
    if ($count == 0) {
    } elsif ($count <= 2) {
       warn "This password has been found in a data breach\n";
    } elsif ($count) {
       die "This password is too insecure\n";
    }

=head2 account

The most common use of the API is to return a list of all breaches a particular account has been involved in. The API takes a single parameter which is the account to be searched for. The account is not case sensitive and will be trimmed of leading or trailing white spaces.

Parameters:

=over 4

=item * truncate - Returns only the name of the breach.

=item * domain - Filters the result set to only breaches against the domain specified. It is possible that one site (and consequently domain), is compromised on multiple occasions.

=item * unverified - Returns breaches that have been flagged as "unverified". By default, only verified breaches are returned web performing a search.

=back 

    use WebService::HIBP();
	use v5.10;

    my $hibp = WebService::HIBP->new();
    foreach my $breach ( $hibp->account('test@example.com', domain => 'adobe.com' ) ) {
	   say $breach->name();
	}

=head2 breach

Sometimes just a single L<breach|WebService::HIBP::Breach> is required and this can be retrieved by the breach L<name|WebService::HIBP::Breach#name>. This is the stable value which may or may not be the same as the breach L<title|WebService::HIBP::Breach#title> (which can change).

=head2 breaches

A L<breach|WebService::HIBP::Breach> is an instance of a system having been compromised by an attacker and the data disclosed. For example, Adobe was a breach, Gawker was a breach etc. This method returns the details of each of breach in the system.

Parameters:

=over 4

=item * domain - Filters the result set to only breaches against the domain specified. It is possible that one site (and consequently domain), is compromised on multiple occasions. 

=back 

=head2 data_classes

A "data class" is an attribute of a record compromised in a breach. For example, many breaches expose data classes such as "Email addresses" and "Passwords". The values returned by this service are ordered alphabetically in a string array and will expand over time as new breaches expose previously unseen classes of data.

=head2 pastes

This method takes a single parameter which is the email address to be searched for. Unlike searching for breaches, usernames that are not email addresses cannot be searched for. The email is not case sensitive and will be trimmed of leading or trailing white spaces.

=head1 DIAGNOSTICS

=over

=item C<< Failed to retrieve %s >>

The URL could not be retrieved. Check network and proxy settings.

=back

=head1 CONFIGURATION AND ENVIRONMENT

WebService::HIBP requires no configuration files or environment variables.  However, it will use the values of C<$ENV{no_proxy}> and C<$ENV{HTTP_PROXY}> as defaults for calls to the L<https://haveibeenpwned.com/api/v2/> API via the LWP::UserAgent module.

=head1 DEPENDENCIES

WebService::HIBP requires the following non-core modules

  JSON
  LWP::UserAgent
  URI::Escape
  Digest::SHA

=head1 INCOMPATIBILITIES

None reported

=head1 BUGS AND LIMITATIONS

Please report any bugs or feature requests to C<bug-webservice-hibp at rt.cpan.org>, or through
the web interface at L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=WebService-HIBP>.  I will be notified, and then you'll
automatically be notified of progress on your bug as I make changes.

=head1 AUTHOR

David Dick, C<< <ddick at cpan.org> >>

=head1 SUPPORT

You can find documentation for this module with the perldoc command.

    perldoc WebService::HIBP


You can also look for information at:

=over 4

=item * RT: CPAN's request tracker (report bugs here)

L<http://rt.cpan.org/NoAuth/Bugs.html?Dist=WebService-HIBP>

=item * AnnoCPAN: Annotated CPAN documentation

L<http://annocpan.org/dist/WebService-HIBP>

=item * CPAN Ratings

L<http://cpanratings.perl.org/d/WebService-HIBP>

=item * Search CPAN

L<http://search.cpan.org/dist/WebService-HIBP/>

=back

=head1 ACKNOWLEDGEMENTS

Thanks to Troy Hunt for providing the service at L<https://haveibeenpwned.com>

POD was extracted from the API help at L<https://haveibeenpwned.com/API/v2>

=head1 LICENSE AND COPYRIGHT

Copyright 2018 David Dick.

This program is free software; you can redistribute it and/or modify it
under the terms of either: the GNU General Public License as published
by the Free Software Foundation; or the Artistic License.

See L<http://dev.perl.org/licenses/> for more information.
